# Copyright (C) 2021 Fraunhofer Institut IOSB, Fraunhoferstr. 1, D 76131
# Karlsruhe, Germany.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import frost_sta_client.model.tasking_capability
from . import entity
from . import tasking_capability

from frost_sta_client.dao.task import TaskDao

from frost_sta_client import utils


class Task(entity.Entity):
    def __init__(self,
                 tasking_parameters=None,
                 creation_time=None,
                 tasking_capability=None,
                 **kwargs):
        super().__init__(**kwargs)
        if tasking_parameters is None:
            tasking_parameters = {}
        self.tasking_parameters = tasking_parameters
        self.creation_time = creation_time
        self.tasking_capability = tasking_capability

    def __new__(cls, *args, **kwargs):
        new_task = super().__new__(cls)
        attributes = {'_id': None, '_tasking_parameters': {}, '_creation_time': None, '_tasking_capability': None,
                      '_self_link': '', '_service': None}
        for key, value in attributes.items():
            new_task.__dict__[key] = value
        return new_task

    @property
    def tasking_parameters(self):
        return self._tasking_parameters

    @tasking_parameters.setter
    def tasking_parameters(self, value):
        if value is None or not isinstance(value, dict):
            raise ValueError('tasking parameter should be of type dict!')
        self._tasking_parameters = value

    @property
    def creation_time(self):
        return self._creation_time

    @creation_time.setter
    def creation_time(self, value):
        self._creation_time = utils.check_datetime(value, 'creation_time')

    @property
    def tasking_capability(self):
        return self._tasking_capability

    @tasking_capability.setter
    def tasking_capability(self, value):
        if value is None:
            self._tasking_capability = None
            return
        if not isinstance(value, tasking_capability.TaskingCapability):
            raise ValueError('tasking capability should be of type TaskingCapability!')
        self._tasking_capability = value

    def ensure_service_on_children(self, service):
        if self.tasking_capability is not None:
            self.tasking_capability.set_service(service)

    def __eq__(self, other):
        if not super().__eq__(other):
            return False
        if self.tasking_parameters != other.tasking_parameters:
            return False
        if self.creation_time != other.creation_time:
            return False
        return True

    def __ne__(self, other):
        return not self == other

    def __getstate__(self):
        data = super().__getstate__()
        if self.tasking_parameters is not None and self.tasking_parameters != {}:
            data['taskingParameters'] = self.tasking_parameters
        if self.creation_time is not None:
            data['creationTime'] = utils.parse_datetime(self.creation_time)
        if self.tasking_capability is not None:
            data['TaskingCapability'] = self.tasking_capability
        return data

    def __setstate__(self, state):
        super().__setstate__(state)
        self.tasking_parameters = state.get('taskingParameters', {})
        self.creation_time = state.get('creationTime', None)
        if state.get('taskingCapability', None) is not None:
            self.tasking_capability = frost_sta_client.model.tasking_capability.TaskingCapability()
            self.tasking_capability.__setstate__(state['taskingCapability'])

    def get_dao(self, service):
        return TaskDao(service)



# Copyright (C) 2021 Fraunhofer Institut IOSB, Fraunhoferstr. 1, D 76131
# Karlsruhe, Germany.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import inspect
import json

import frost_sta_client.model.ext.unitofmeasurement
from frost_sta_client.dao.datastream import DatastreamDao

from . import thing
from . import sensor
from . import observedproperty
from . import observation
from . import entity
from .ext import unitofmeasurement
from .ext import entity_list
from .ext import entity_type

from frost_sta_client import utils

import geojson.geometry


class Datastream(entity.Entity):

    def __init__(self,
                 name='',
                 description='',
                 observation_type='',
                 unit_of_measurement=None,
                 observed_area=None,
                 properties=None,
                 phenomenon_time=None,
                 result_time=None,
                 thing=None,
                 sensor=None,
                 observed_property=None,
                 observations=None,
                 **kwargs):
        """
        This class handles Datastreams assigned to a Thing. Before you create a Datastreams, you firstly have to
        create a Thing, a Sensor and an observedProperty to which you have to refer by specifying its ids.

        Parameters
        ----------
        name: str
        description: str
        observation_type: str
        unit_of_measurement: dict
            Should be a dict of keys 'name', 'symbol', 'definition' with values of str.

        """
        super().__init__(**kwargs)
        if properties is None:
            properties = {}
        self.name = name
        self.description = description
        self.observation_type = observation_type
        self.unit_of_measurement = unit_of_measurement
        self.observed_area = observed_area
        self.properties = properties
        self.phenomenon_time = phenomenon_time
        self.result_time = result_time
        self.thing = thing
        self.sensor = sensor
        self.observations = observations
        self.observed_property = observed_property


    def __new__(cls, *args, **kwargs):
        new_datastream = super().__new__(cls)
        attributes = dict(_id=None, _name='', _description='', _properties={}, _observation_type='',
                          _unit_of_measurement=None, _observed_area=None, _phenomenon_time=None, _result_time=None,
                          _thing=None, _sensor=None, _observed_property=None, _observations=None, _self_link='',
                          _service=None)
        for key, value in attributes.items():
            new_datastream.__dict__[key] = value
        return new_datastream

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        if value is None:
            self._name = None
            return
        if not isinstance(value, str):
            raise ValueError('name should be of type str!')
        self._name = value

    @property
    def description(self):
        return self._description

    @description.setter
    def description(self, value):
        if value is None:
            self._description = None
            return
        if not isinstance(value, str):
            raise ValueError('description should be of type str!')
        self._description = value

    @property
    def observation_type(self):
        return self._observation_type

    @observation_type.setter
    def observation_type(self, value):
        if value is None:
            self._observation_type = None
            return
        if not isinstance(value, str):
            raise ValueError('observation_type should be of type str!')
        self._observation_type = value

    @property
    def unit_of_measurement(self):
        return self._unit_of_measurement

    @unit_of_measurement.setter
    def unit_of_measurement(self, value):
        if value is None or isinstance(value, unitofmeasurement.UnitOfMeasurement):
            self._unit_of_measurement = value
            return
        raise ValueError('unitOfMeasurement should be of type UnitOfMeasurement!')

    @property
    def observed_area(self):
        return self._observed_area

    @observed_area.setter
    def observed_area(self, value):
        if value is None:
            self._location = None
            return
        geo_classes = [obj for _, obj in inspect.getmembers(geojson) if inspect.isclass(obj) and
                       obj.__module__ == 'geojson.geometry']
        if type(value) in geo_classes:
            self._observed_area = value
            return
        else:
            try:
                json.dumps(value)
            except TypeError:
                raise ValueError('observedArea should be of json_serializable!')
            self._observed_area = value

    @property
    def properties(self):
        return self._properties

    @properties.setter
    def properties(self, value):
        if value is None:
            self._properties = None
            return
        if not isinstance(value, dict):
            raise ValueError('properties should be of type dict')
        self._properties = value

    @property
    def phenomenon_time(self):
        return self._phenomenon_time

    @phenomenon_time.setter
    def phenomenon_time(self, value):
        self._phenomenon_time = utils.check_datetime(value, 'phenomenon_time')

    @property
    def result_time(self):
        return self._result_time

    @result_time.setter
    def result_time(self, value):
        self._result_time = utils.check_datetime(value, 'result_time')

    @property
    def thing(self):
        return self._thing

    @thing.setter
    def thing(self, value):
        if value is None or isinstance(value, thing.Thing):
            self._thing = value
            return
        raise ValueError('thing should be of type Thing!')

    @property
    def sensor(self):
        return self._sensor

    @sensor.setter
    def sensor(self, value):
        if value is None or isinstance(value, sensor.Sensor):
            self._sensor = value
            return
        raise ValueError('sensor should be of type Sensor!')

    @property
    def observed_property(self):
        return self._observed_property

    @observed_property.setter
    def observed_property(self, value):
        if isinstance(value, observedproperty.ObservedProperty) or value is None:
            self._observed_property = value
            return
        raise ValueError('observed property should by of type ObservedProperty!')

    @property
    def observations(self):
        return self._observations

    @observations.setter
    def observations(self, values):
        if values is None:
            self._observations = None
            return
        if isinstance(values, list) and all(isinstance(ob, observation.Observation) for ob in values):
            entity_class = entity_type.EntityTypes['Observation']['class']
            self._observations = entity_list.EntityList(entity_class=entity_class, entities=values)
            return
        if isinstance(values, entity_list.EntityList) and \
                all(isinstance(ob, observation.Observation) for ob in values.entities):
            self._observations = values
            return
        raise ValueError('Observations should be a list of Observations')

    def get_observations(self):
        result = self.service.observations()
        result.parent = self
        return result

    def ensure_service_on_children(self, service):
        if self.thing is not None:
            self.thing.set_service(service)
        if self.sensor is not None:
            self.sensor.set_service(service)
        if self.observed_property is not None:
            self.observed_property.set_service(service)
        if self.observations is not None:
            self.observations.set_service(service)

    def __eq__(self, other):
        if not super().__eq__(other):
            return False
        if self.name != other.name:
            return False
        if self.description != other.description:
            return False
        if self.observation_type != other.observation_type:
            return False
        if self.unit_of_measurement != other.unit_of_measurement:
            return False
        if self.properties != other.properties:
            return False
        if self.result_time != other.result_time:
            return False
        return True

    def __ne__(self, other):
        return not self == other

    def __getstate__(self):
        data = super().__getstate__()
        if self.name is not None and self.name != '':
            data['name'] = self.name
        if self.description is not None and self.description != '':
            data['description'] = self.description
        if self.observation_type is not None and self.observation_type != '':
            data['observationType'] = self.observation_type
        if self.properties is not None and self.properties != {}:
            data['properties'] = self.properties
        if self.unit_of_measurement is not None:
            data['unitOfMeasurement'] = self.unit_of_measurement
        if self.observed_area is not None:
            data['observedArea'] = self.observed_area
        if self.phenomenon_time is not None:
            data['phenomenonTime'] = utils.parse_datetime(self.phenomenon_time)
        if self.result_time is not None:
            data['resultTime'] = utils.parse_datetime(self.result_time)
        if self.thing is not None:
            data['Thing'] = self.thing
        if self.sensor is not None:
            data['Sensor'] = self.sensor
        if self.observed_property is not None:
            data['ObservedProperty'] = self.observed_property
        if self.observations is not None and len(self.observations.entities) > 0:
            data['Observations'] = self.observations.__getstate__()
        return data

    def __setstate__(self, state):
        super().__setstate__(state)
        self.name = state.get("name", None)
        self.description = state.get("description", None)
        self.observation_type = state.get("observationType", None)
        self.properties = state.get("properties", {})
        if state.get("unitOfMeasurement", None) is not None:
            self.unit_of_measurement = frost_sta_client.model.ext.unitofmeasurement.UnitOfMeasurement()
            self.unit_of_measurement.__setstate__(state["unitOfMeasurement"])
        if state.get("observedArea", None) is not None:
            self.observed_area = frost_sta_client.utils.process_area(state["observedArea"])
        if state.get("phenomenonTime", None) is not None:
            self.phenomenon_time = state["phenomenonTime"]
        if state.get("resultTime", None) is not None:
            self.result_time = state["resultTime"]
        if state.get("Thing", None) is not None:
            self.thing = frost_sta_client.model.thing.Thing()
            self.thing.__setstate__(state["Thing"])
        if state.get("ObservedProperty", None) is not None:
            self.observed_property = frost_sta_client.model.observedproperty.ObservedProperty()
            self.observed_property.__setstate__(state["ObservedProperty"])
        if state.get("Sensor", None) is not None:
            self.sensor = frost_sta_client.model.sensor.Sensor()
            self.sensor.__setstate__(state["Sensor"])
        if state.get("Observations", None) is not None and isinstance(state["Observations"], list):
            entity_class = entity_type.EntityTypes['Observation']['class']
            self.observations = utils.transform_json_to_entity_list(state['Observations'], entity_class)
            self.observations.next_link = state.get("Observations@iot.nextLink", None)
            self.observations.count = state.get("Observations@iot.count", None)

    def get_dao(self, service):
        return DatastreamDao(service)



__all__ = ['actuator', 'base', 'datastream', 'features_of_interest', 'historical_location', 'location',
           'multi_datastream', 'observation', 'observedproperty', 'sensor', 'task', 'tasking_capability', 'thing']



import unittest

import frost_sta_client.model.observation
import frost_sta_client.utils
import frost_sta_client.model.ext.entity_type
import datetime


class TestEntityReader(unittest.TestCase):

    def test_write_thing_basic_success(self):
        json_dict = {
            'phenomenonTime': '2016-01-07T02:00:00.000+00:00',
            'resultTime': None,
            'result': '0.15',
            'Datastream@iot.navigationLink': 'https://server.de/SensorThingsService/v1.0/'
                                             'Observations(7179373)/Datastream',
            'FeatureOfInterest@iot.navigationLink': 'https://server.de/SensorThingsService/v1.0/'
                                                    'Observations(7179373)/FeatureOfInterest',
            '@iot.id': '719373',
            '@iot.selfLink': 'https://server.de/SensorThingsService/v1.0/Observations(7179373)'
        }
        result = frost_sta_client.utils.transform_json_to_entity(json_dict, "frost_sta_client.model.observation.Observation")

        exp_observation = frost_sta_client.model.observation.Observation(
            phenomenon_time='2016-01-07T02:00:00.000+00:00',
            result='0.15',
            id='719373',
            self_link='https://server.de/SensorThingsService/v1.0/Observations(7179373)'
        )
        self.assertEqual(result, exp_observation)

    def test_read_entity_list(self):
        json_dict = {
            "@iot.nextLink": "https://server.de/SensorThingsService/v1.0/Things?$top=2&"
                             "$skip=14&$expand=Datastreams%28%24top%3D2%3B%24count%3Dtrue%29",
            "value": [
                {
                    "name": "Recoaro 1000",
                    "description": "Weather station Recoaro 1000",
                    "Datastreams@iot.navigationLink": "https://server.de/SensorThingsService/v1.0/Things(19)/"
                                                      "Datastreams",
                    "Datastreams@iot.count": 6,
                    "Datastreams": [
                        {
                            "name": "Air Temperature Recoaro 1000",
                            "description": "The Air Temperature at Recoaro 1000",
                            "observationType": "http://www.opengis.net/def/observationType/OGC-OM/2.0/OM_Measurement",
                            "phenomenonTime": "2010-01-01T00:00:00.000Z/2019-01-13T06:00:00.000Z",
                            "unitOfMeasurement": {
                                "name": "degree celcius",
                                "symbol": "°C",
                                "definition": "ucum:Cel",
                            },
                            "Observations": [
                                {
                                    "result": 0.0
                                }
                            ],
                            "@iot.id": 66,
                            "@iot.selfLink": "https://server.de/SensorThingsService/v1.0/Datastreams(66)",
                        },
                        {
                            "name": "Precipitation Recoaro 1000",
                            "description": "The Precipitation at Recoaro 1000",
                            "observationType": "http://www.opengis.net/def/observationType/OGC-OM/2.0/OM_Measurement",
                            "phenomenonTime": "2010-01-01T00:00:00.000Z/2019-01-13T06:00:00.000Z",
                            "unitOfMeasurement": {
                                "name": "mm/h",
                                "symbol": "mm/h",
                                "definition": "ucum:mm/h"
                            },
                            "@iot.id": 130,
                            "@iot.selfLink": "https://server.de/SensorThingsService/v1.0/Datastreams(130)"
                        }
                    ],
                    "Datastreams@iot.nextLink": "https://server.de/SensorThingsService/v1.0/Things(19)/"
                                                "Datastreams?$top=2&$skip=2&$count=true",
                    "MultiDatastreams@iot.navigationLink": "https://server.de/SensorThingsService/v1.0/"
                                                           "Things(19)/MultiDatastreams",
                    "Locations@iot.navigationLink": "https://server.de/SensorThingsService/v1.0/Things(19)/Locations",
                    "HistoricalLocations@iot.navigationLink": "https://server.de/SensorThingsService/v1.0/"
                                                              "Things(19)/HistoricalLocations",
                    "@iot.id": 19,
                    "@iot.selfLink": "https://server.de/SensorThingsService/v1.0/Things(19)"
                },
                {
                    "name": "Valdagno",
                    "description": "Weather station Valdagno",
                    "Datastreams@iot.navigationLink": "https://server.de/SensorThingsService/v1.0/"
                                                      "Things(20)/Datastreams",
                    "Datastreams": [],
                    "Datastreams@iot.count": 6,
                    "Datastreams@iot.nextLink": "https://server.de/SensorThingsService/v1.0/Things(20)/"
                                                "Datastreams?$top=2&$skip=2&$count=true",
                    "MultiDatastreams@iot.navigationLink": "https://server.de/SensorThingsService/v1.0/"
                                                           "Things(20)/MultiDatastreams",
                    "Locations@iot.navigationLink": "https://server.de/SensorThingsService/v1.0/Things(20)/Locations",
                    "HistoricalLocations@iot.navigationLink": "https://server.de/SensorThingsService/v1.0/"
                                                              "Things(20)/HistoricalLocations",
                    "@iot.id": 20,
                    "@iot.selfLink": "https://server.de/SensorThingsService/v1.0/Things(20)"
                }
            ]
        }

        things = frost_sta_client.utils.transform_json_to_entity_list(json_dict, 'frost_sta_client.model.thing.Thing')

        self.assertEqual('https://server.de/SensorThingsService/v1.0/Things?$top=2&$'
                         'skip=14&$expand=Datastreams%28%24top%3D2%3B%24count%3Dtrue%29', things.next_link)

        thing = things.entities[0]
        self.assertEqual(19, thing.id)
        self.assertEqual('Recoaro 1000', thing.name)
        self.assertEqual('Weather station Recoaro 1000', thing.description)

        ds_list = thing.datastreams
        self.assertEqual('https://server.de/SensorThingsService/v1.0/Things(19)/Datastreams?$top=2&$skip=2&$count=true',
                         ds_list.next_link)
        # self.assertEqual(6, ds_list.count)

        ds = ds_list.entities[0]
        self.assertEqual('Air Temperature Recoaro 1000', ds.name)
        self.assertEqual('The Air Temperature at Recoaro 1000', ds.description)

        result = ds.observations.entities[0].result
        self.assertEqual(0, result)

        ds = ds_list.entities[1]
        self.assertEqual(130, ds.id)
        self.assertEqual('Precipitation Recoaro 1000', ds.name)
        self.assertEqual('The Precipitation at Recoaro 1000', ds.description)

        thing = things.entities[1]
        self.assertEqual(20, thing.id)
        self.assertEqual('Valdagno', thing.name)
        self.assertEqual('Weather station Valdagno', thing.description)

        ds_list = thing.datastreams
        self.assertEqual('https://server.de/SensorThingsService/v1.0/Things(20)/Datastreams?$top=2&$skip=2&$count=true',
                         ds_list.next_link)
        # self.assertEqual(6, ds_list.count)

    def test_read_empty_entity_list(self):
        json_dict = {'value': []}
        things = frost_sta_client.utils.transform_json_to_entity_list(json_dict, 'frost_sta_client.model.thing.Thing')

        self.assertEqual(None, things.next_link)
        self.assertTrue(len(things.entities) == 0)

    def test_read_tasking_capabilities(self):
        json_dict = {
            'name': 'createNewVA',
            'description': 'Virtual Actuator Server, starts new Virtual Actuators',
            'taskingParameters': {
                'type': 'DataRecord',
                'field': [{
                    'type': 'Text',
                    'label': 'Aktor-Name',
                    'description': 'Name des neuen virtuellen Aktors',
                    'name': 'vaName'
                }, {
                    'type': 'Text',
                    'label': 'Aktor-Beschreibung',
                    'description': 'Beschreibung des neuen virtuellen Aktors',
                    'name': 'vaDescription'
                }]
            },
            'Actuator@iot.navigationLink': 'https://server.de/SensorThingsService/v1.0/Actuator',
            'Thing@iot.navigationLink': 'https://server.de/SensorThingsService/v1.0/TaskingCapabilities(1)/Thing',
            'Tasks@iot.navigationLink': 'https://server.de/SensorThingsService/v1.0/TaskingCapabilities(1)/Tasks',
            '@iot.id': 1,
            '@iot.selfLink': 'https://server.de/SensorThingsService/v1.0/TaskingCapabilities(1)'}

        capability = frost_sta_client.utils. \
            transform_json_to_entity(json_dict,
                                     'frost_sta_client.model.tasking_capability.TaskingCapability')

        expected = frost_sta_client.model.tasking_capability.TaskingCapability()
        expected.name = 'createNewVA'
        expected.description = 'Virtual Actuator Server, starts new Virtual Actuators'
        expected.tasking_parameters = {
            'type': 'DataRecord',
            'field': [{
                'type': 'Text',
                'label': 'Aktor-Name',
                'description': 'Name des neuen virtuellen Aktors',
                'name': 'vaName'
            }, {
                'type': 'Text',
                'label': 'Aktor-Beschreibung',
                'description': 'Beschreibung des neuen virtuellen Aktors',
                'name': 'vaDescription'
            }]
        }
        expected.id = 1
        expected.self_link = 'https://server.de/SensorThingsService/v1.0/TaskingCapabilities(1)'

        self.assertEqual(expected, capability)

    def test_read_tasking_capabilities_with_constraint(self):
        json_dict = {
            'name': 'DatastreamCopierCapability',
            'description': 'Copies Observations from one Datastream to another',
            'taskingParameters':
                {
                    'type': 'DataRecord',
                    'field': [{
                        'type': 'Count',
                        'name': 'sourceDatastream',
                        'label': 'Source Datastream',
                        'description': 'ID of the datastream from which the observations should be taken.',
                        'constraint': {
                            'type': 'AllowedValues',
                            'interval': [[0, 10000]]
                        }
                    },
                        {
                            'type': 'Count',
                            'name': 'destinationDatastream',
                            'label': 'Destination Datastream',
                            'description': 'ID of the datastream to which the observations should be copied.',
                            'constraint': {
                                'type': 'AllowedValues',
                                'interval': [[0, 10000]]
                            }
                        }]
                },
            'Actuator@iot.navigationLink': 'https://server.de/SensorThingsService/v1.0/Actuator',
            'Thing@iot.navigationLink': 'https://server.de/SensorThingsService/v1.0/TaskingCapabilities(1)/Thing',
            'Tasks@iot.navigationLink': 'https://server.de/SensorThingsService/v1.0/TaskingCapabilities(1)/Tasks',
            '@iot.id': 1,
            '@iot.selfLink': 'https://server.de/SensorThingsService/v1.0/TaskingCapabilities(1)'
        }

        entity_class = 'frost_sta_client.model.tasking_capability.TaskingCapability'
        capability = frost_sta_client.utils.transform_json_to_entity(json_dict, entity_class)

        expected = frost_sta_client.model.tasking_capability.TaskingCapability()
        expected.name = 'DatastreamCopierCapability'
        expected.description = 'Copies Observations from one Datastream to another'
        expected.id = 1
        expected.tasking_parameters = {
                    'type': 'DataRecord',
                    'field': [{
                        'type': 'Count',
                        'name': 'sourceDatastream',
                        'label': 'Source Datastream',
                        'description': 'ID of the datastream from which the observations should be taken.',
                        'constraint': {
                            'type': 'AllowedValues',
                            'interval': [[0, 10000]]
                        }
                    },
                        {
                            'type': 'Count',
                            'name': 'destinationDatastream',
                            'label': 'Destination Datastream',
                            'description': 'ID of the datastream to which the observations should be copied.',
                            'constraint': {
                                'type': 'AllowedValues',
                                'interval': [[0, 10000]]
                            }
                        }]
                }

        self.assertEqual(expected, capability)



# Copyright (C) 2021 Fraunhofer Institut IOSB, Fraunhoferstr. 1, D 76131
# Karlsruhe, Germany.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from requests.auth import HTTPBasicAuth


class AuthHandler:
    def __init__(self, username="", password=""):
        self.username = username
        self.password = password

    def add_auth_header(self):
        if not (self.username is None or self.password is None):
            return HTTPBasicAuth(self.username, self.password)



# Copyright (C) 2021 Fraunhofer Institut IOSB, Fraunhoferstr. 1, D 76131
# Karlsruhe, Germany.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from . import entity

import frost_sta_client.dao.features_of_interest

from .ext import entity_list, entity_type

import json

from .. import utils


class FeatureOfInterest(entity.Entity):
    def __init__(self,
                 name='',
                 description='',
                 encoding_type='',
                 feature=None,
                 properties=None,
                 observations=None,
                 **kwargs):
        super().__init__(**kwargs)
        if properties is None:
            properties = {}
        self.name = name
        self.description = description
        self.encoding_type = encoding_type
        self.feature = feature
        self.properties = properties
        self.observations = observations

    def __new__(cls, *args, **kwargs):
        new_foi = super().__new__(cls)
        attributes = {'_id': None, '_name': '', '_description': '', '_properties': {}, '_encoding_type': '',
                      '_feature': '', '_observations': None, '_self_link': '', '_service': None}
        for key, value in attributes.items():
            new_foi.__dict__[key] = value
        return new_foi

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        if value is None:
            self._name = None
            return
        if not isinstance(value, str):
            raise ValueError('name should be of type str!')
        self._name = value

    @property
    def description(self):
        return self._description

    @description.setter
    def description(self, value):
        if value is None:
            self._description = None
            return
        if not isinstance(value, str):
            raise ValueError('description should be of type str!')
        self._description = value

    @property
    def properties(self):
        return self._properties

    @properties.setter
    def properties(self, value):
        if value is None:
            self._properties = None
            return
        if not isinstance(value, dict):
            raise ValueError('properties should be of type dict!')
        self._properties = value

    @property
    def encoding_type(self):
        return self._encoding_type

    @encoding_type.setter
    def encoding_type(self, value):
        if value is None:
            self._encoding_type = None
            return
        if not isinstance(value, str):
            raise ValueError('encodingType should be of type str!')
        self._encoding_type = value

    @property
    def observations(self):
        return self._observations

    @observations.setter
    def observations(self, values):
        if values is None:
            self._observations = None
            return
        if isinstance(values, list) and \
                all(isinstance(ob, frost_sta_client.model.observation.Observation) for ob in values):
            entity_class = entity_type.EntityTypes['Observation']['class']
            self._observations = entity_list.EntityList(entity_class=entity_class, entities=values)
            return
        if isinstance(values, entity_list.EntityList) and \
                all((isinstance(ob, frost_sta_client.model.observation.Observation)) for ob in values.entities):
            self._observations = values
            return
        raise ValueError('Observations should be a list of Observations')

    @property
    def feature(self):
        return self._feature

    @feature.setter
    def feature(self, value):
        if value is None:
            self._feature = None
            return
        try:
            json.dumps(value)
        except TypeError:
            raise TypeError('feature should be json serializable')
        self._feature = value

    def get_observations(self):
        result = self.service.observations()
        result.parent = self
        return result

    def ensure_service_on_children(self, service):
        if self.observations is not None:
            self.observations.set_service(service)

    def __eq__(self, other):
        if not super().__eq__(other):
            return False
        if self.name != other.name:
            return False
        if self.properties != other.properties:
            return False
        if self.description != other.description:
            return False
        if self.encoding_type != other.encoding_type:
            return False
        if self.feature != other.feature:
            return False
        return True

    def __ne__(self, other):
        return not self == other

    def __getstate__(self):
        data = super().__getstate__()
        if self.name is not None and self.name != '':
            data['name'] = self.name
        if self.description is not None and self.description != '':
            data['description'] = self.description
        if self.properties is not None and self.properties != {}:
            data['properties'] = self.properties
        if self.encoding_type is not None and self.encoding_type != '':
            data['encodingType'] = self.encoding_type
        if self.feature is not None:
            data['feature'] = self.feature
        if self.observations is not None and len(self.observations.entities) > 0:
            data['Observations'] = self.observations.__gestate__()
        return data

    def __setstate__(self, state):
        super().__setstate__(state)
        self.name = state.get("name", None)
        self.description = state.get("description", None)
        self.properties = state.get("properties", {})
        self.encoding_type = state.get("encodingType", None)
        self.feature = state.get("feature", None)
        if state.get("Observations", None) is not None and isinstance(state["Observations"], list):
            entity_class = entity_type.EntityTypes['Observation']['class']
            self.observations = utils.transform_json_to_entity_list(state['Observations'], entity_class)
            self.observations.next_link = state.get("Observations@iot.nextLink", None)
            self.observations.count = state.get("Observations@iot.count", None)

    def get_dao(self, service):
        return frost_sta_client.dao.features_of_interest.FeaturesOfInterestDao(service)
